/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotSensorMemory
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD
#include <atomic>
#include <thread>
#include <queue>

// Simox
#include <SimoxUtility/meta/enum/adapt_enum.h>

// ArmarX
#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/services/tasks/PeriodicTask.h>

// BaseClass
#include <RobotAPI/libraries/RobotAPIComponentPlugins/RobotUnitComponentPlugin.h>
#include "RobotAPI/libraries/RobotAPIComponentPlugins/ArVizComponentPlugin.h"

#include <RobotAPI/libraries/RobotUnitDataStreamingReceiver/RobotUnitDataStreamingReceiver.h>
#include <RobotAPI/libraries/armem/server/ComponentPlugin.h>

#include <RobotAPI/libraries/armem_robot_state/server/description/Segment.h>
#include <RobotAPI/libraries/armem_robot_state/server/proprioception/Segment.h>
#include <RobotAPI/libraries/armem_robot_state/server/localization/Segment.h>
#include <RobotAPI/libraries/armem_robot_state/server/common/Visu.h>

namespace armarx::armem::server::robot_state
{

    /**
     * @defgroup Component-RobotSensorMemory RobotSensorMemory
     * @ingroup RobotAPI-Components
     * A description of the component RobotSensorMemory.
     *
     * @class RobotSensorMemory
     * @ingroup Component-RobotSensorMemory
     * @brief Brief description of class RobotSensorMemory.
     *
     * Detailed description of class RobotSensorMemory.
     */
    class RobotStateMemory :
        virtual public armarx::Component,
        virtual public armem::server::ComponentPluginUser,
        virtual public RobotUnitComponentPluginUser,
        virtual public armarx::ArVizComponentPluginUser
    {
    public:
        RobotStateMemory();

        virtual ~RobotStateMemory();

        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;

    protected:
        /// @see armarx::ManagedIceObject::onInitComponent()
        void onInitComponent() override;

        /// @see armarx::ManagedIceObject::onConnectComponent()
        void onConnectComponent() override;

        /// @see armarx::ManagedIceObject::onDisconnectComponent()
        void onDisconnectComponent() override;

        /// @see armarx::ManagedIceObject::onExitComponent()
        void onExitComponent() override;

        /// @see PropertyUser::createPropertyDefinitions()
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

    private:
        // RobotUnit Streaming
        void convertRobotUnitStreamingDataToAron();
        void storeConvertedRobotUnitDataInMemory();

        void startRobotUnitStream();
        void stopRobotUnitStream();

    private:
        std::string workingMemoryName = "RobotStateMemory";
        bool addCoreSegmentOnUsage = false;

        mutable std::recursive_mutex startStopMutex;

        // Core segments
        // - description
        description::Segment descriptionSegment;

        // - proprioception
        proprioception::Segment proprioceptionSegment;
        armem::data::MemoryID robotUnitProviderID;

        // - localization
        localization::Segment localizationSegment;

        // Joint visu for all segments => robot pose and configuration
        Visu commonVisu;

        // RobotUnit stuff
        RobotUnitDataStreaming::DataStreamingDescription keys;
        std::vector<RobotUnitDataStreaming::DataEntry> keysList;
        RobotUnitDataStreaming::Config cfg;
        RobotUnitDataStreamingReceiverPtr handler;
        std::map<std::string, std::string> configSensorMapping;

        struct RobotUnitData
        {
            struct RobotUnitDataGroup
            {
                long timestamp;
                std::string name;
                aron::datanavigator::DictNavigatorPtr data;
            };

            std::map<std::string, RobotUnitDataGroup> groups;
        };

        // params
        static constexpr int ROBOT_UNIT_MAXIMUM_FREQUENCY = 100;
        static constexpr const char* sensorValuePrefix = "RobotUnit.";

        int robotUnitPollFrequency = 50;
        std::string robotUnitSensorPrefix = "sens.*";
        unsigned int robotUnitMemoryBatchSize = 50;
        std::string robotUnitConfigPath = "NO CONFIG SET";

        // queue
        std::queue<RobotUnitData> robotUnitDataQueue;
        mutable std::mutex robotUnitDataMutex;

        // running tasks
        armarx::PeriodicTask<RobotStateMemory>::pointer_type robotUnitConversionTask;
        armarx::PeriodicTask<RobotStateMemory>::pointer_type robotUnitStoringTask;
    };

}  // namespace armarx::armem::server::robot_state
