#pragma once

#include "../base/EntityBase.h"

#include "EntitySnapshot.h"

#include "../workingmemory/Entity.h"


namespace armarx::armem::ltm
{
    /**
     * @brief An entity over a period of time.
     *
     * An entity should be a physical thing or abstract concept existing
     * (and potentially evolving) over some time.
     *
     * Examples are:
     * - objects (the green box)
     * - agents (robot, human)
     * - locations (frige, sink)
     * - grasp affordances (general, or for a specific object)
     * - images
     * - point clouds
     * - other sensory values
     *
     * At each point in time (`EntitySnapshot`), the entity can have a
     * (potentially variable) number of instances (`EntityInstance`),
     * each containing a single `AronData` object of a specific `AronType`.
     */
    class Entity :
        public base::EntityBase<EntitySnapshot, Entity>
    {
        using Base = base::EntityBase<EntitySnapshot, Entity>;

    public:

        using Base::EntityBase;
        using Base::operator=;

        Entity()
        {
            // the history of snapshots is just a cache of frequently used elements
            setMaxHistorySize(MAX_HISTORY_SIZE);
        }

        // Conversion
        wm::Entity convert() const;

        // MongoDB connection
        void reload();
        void append(const wm::Entity&);

        // virtual overrides for LTM lookups
        virtual bool hasSnapshot(const Time& time) const override;
        virtual std::vector<Time> getTimestamps() const override;
        virtual const EntitySnapshot& getSnapshot(const Time& time) const override;

    protected:
        virtual void _copySelf(Entity& other) const override
        {
            Base::_copySelf(other);
            other.dbsettings = dbsettings;
        }

        virtual void _copySelfEmpty(Entity& other) const override
        {
            Base::_copySelfEmpty(other);
            other.dbsettings = dbsettings;
        }

    protected:
        // virtual overrides for LTM storage
        virtual const std::map<Time, EntitySnapshot>::value_type& getLatestItem() const override;

    public:
        MongoDBConnectionManager::MongoDBSettings dbsettings;

    private:
        const static constexpr int MAX_HISTORY_SIZE = 200;
    };

}
