/*
* This file is part of ArmarX.
*
* Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
* Karlsruhe Institute of Technology (KIT), all rights reserved.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL

// Header
#include <RobotAPI/libraries/aron/core/codegenerator/codeWriter/cpp/serializer/Serializer.h>

// ArmarX
#include <RobotAPI/libraries/aron/core/codegenerator/codeWriter/cpp/serializer/SerializerFactory.h>

namespace armarx::aron::cppserializer
{
    // constantes
    const std::string Serializer::ARON_DATA_NAME = simox::meta::get_type_name(typeid(armarx::aron::data::AronData));
    const std::string Serializer::ARON_DATA_PTR_NAME = Serializer::ARON_DATA_NAME + "::PointerType";

    const std::string Serializer::READ_START_RETURN_TYPE_ACCESSOR = "_return_type";

    const std::string Serializer::ARON_TYPE_NAME = simox::meta::get_type_name(typeid(armarx::aron::type::AronType));
    const std::string Serializer::ARON_TYPE_PTR_NAME = Serializer::ARON_TYPE_NAME + "::PointerType";

    const std::map<std::string, std::string> Serializer::ESCAPE_ACCESSORS =
    {
        {"->", "__ptr__"},
        {".", "__dot__"},
        {"[", "__lbrC__"},
        {"]", "__rbrC__"},
        {"(", "__lbrR__"},
        {")", "__rbrR__"},
        {"*", "__ast__"},
    };

    const SerializerFactoryPtr Serializer::FACTORY = SerializerFactoryPtr(new SerializerFactory());


    // constructors
    Serializer::Serializer(const std::string& cppName, const std::string& aronDataTypename, const std::string& aronTypeTypename) :
        cppTypename(cppName),
        aronDataTypename(aronDataTypename),
        aronTypeTypename(aronTypeTypename)
    {

    }

    // static methods
    std::string Serializer::ResolveMaybeAccessor(const std::string& s, const typenavigator::NavigatorPtr& t)
    {
        if (t->getMaybe() == type::Maybe::eOptional)
        {
            return s + ".value()";
        }
        if (t->getMaybe() == type::Maybe::eRawPointer || t->getMaybe() == type::Maybe::eSharedPointer || t->getMaybe() == type::Maybe::eUniquePointer)
        {
            return "*" + s;
        }
        return s;
    }

    CppBlockPtr Serializer::ResolveMaybeWriteBlock(const std::string& accessor, const CppBlockPtr& block_if_data, const typenavigator::NavigatorPtr& typenavigator)
    {
        CppBlockPtr b = std::make_shared<CppBlock>();

        if (!accessor.empty())
        {
            b->addLine("// Generate WriteInterface API calls for " + accessor);
        }
        else
        {
            b->addLine("// Generate WriteInterface API calls for the top level object " + ExtractCppTypename(typenavigator));
        }

        if (typenavigator->getMaybe() != type::Maybe::eNone)
        {
            switch (typenavigator->getMaybe())
            {
                case type::Maybe::eOptional:
                {
                    b->addLine("if (" + accessor + ".has_value())");
                    b->addBlock(block_if_data);
                    break;
                }
                case type::Maybe::eRawPointer:
                {
                    b->addLine("if (" + accessor + " != NULL)");
                    b->addBlock(block_if_data);
                    break;
                }
                case type::Maybe::eSharedPointer:
                    [[fallthrough]];
                case type::Maybe::eUniquePointer:
                {
                    b->addLine("if (" + accessor + " != nullptr)");
                    b->addBlock(block_if_data);
                    break;
                }
                default:
                {
                    throw error::MaybeNotValidException("Serializer", "ResolveMaybeWriteBlock", "Unkown Maybe type (perhaps forgot to add in switch case?)", typenavigator->getMaybe(), typenavigator->getPath());
                }
            }

            b->addLine("else");
            CppBlockPtr block_if_no_data = std::make_shared<CppBlock>();
            block_if_no_data->addLine("w.writeNull(); // The maybe type '" + accessor + "' has no data, so we simply write NULL (resulting in a AronData nullhandle)");
            b->addBlock(block_if_no_data);
            return b;
        }
        b->appendBlock(block_if_data);
        return b;
    }

    CppBlockPtr Serializer::ResolveMaybeReadBlock(const std::string& accessor, const std::string& readElement, const CppBlockPtr& block_if_data, const typenavigator::NavigatorPtr& typenavigator)
    {
        std::string escaped_accessor = EscapeAccessor(accessor);
        std::string read_start_result_accessor = escaped_accessor + READ_START_RETURN_TYPE_ACCESSOR;

        CppBlockPtr resolved_block = std::make_shared<CppBlock>();

        if (!accessor.empty())
        {
            resolved_block->addLine("// Generate ReadInterface API calls for " + accessor);
        }
        else
        {
            resolved_block->addLine("// Generate ReadInterface API calls for the top level object " + ExtractCppTypename(typenavigator));
        }

        if (!readElement.empty())
        {
            resolved_block->addLine("auto " + read_start_result_accessor + " = " + readElement + "; // of " + accessor);
        }

        if (typenavigator->getMaybe() != type::Maybe::eNone)
        {
            resolved_block->addLine("if (" + read_start_result_accessor + ".success)");
            CppBlockPtr if_success = std::make_shared<CppBlock>();

            switch (typenavigator->getMaybe())
            {
                case type::Maybe::eOptional:
                {
                    auto s = FromAronTypeNaviagtorPtr(typenavigator);
                    if_success->addLine(accessor + " = " + s->getFullTypenameGenerator() + "();");
                    if_success->appendBlock(block_if_data);
                    break;
                }
                case type::Maybe::eRawPointer:
                {
                    auto s = FromAronTypeNaviagtorPtr(typenavigator);
                    if_success->addLine(accessor + " = " + s->getFullTypenameGenerator() + "();");
                    if_success->appendBlock(block_if_data);
                    break;
                }
                case type::Maybe::eSharedPointer:
                    [[fallthrough]];
                case type::Maybe::eUniquePointer:
                {
                    auto s = FromAronTypeNaviagtorPtr(typenavigator);
                    if_success->addLine(accessor + " = " + s->getFullTypenameGenerator() + "();");
                    if_success->appendBlock(block_if_data);
                    break;
                }
                default:
                {
                    throw error::MaybeNotValidException("Serializer", "ResolveMaybeReadBlock", "Unkown Maybe type (perhaps forgot to add in switch case?)", typenavigator->getMaybe(), typenavigator->getPath());
                }
            }
            resolved_block->addBlock(if_success);
            return resolved_block;
        }
        resolved_block->appendBlock(block_if_data);
        return resolved_block;
    }

    CppBlockPtr Serializer::ResolveMaybeEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor, const CppBlockPtr& block_if_data, const typenavigator::NavigatorPtr& typenavigator)
    {
        CppBlockPtr b = std::make_shared<CppBlock>();
        if (!accessor.empty())
        {
            b->addLine("// Comparing " + accessor + " and " + otherInstanceAccessor);
        }
        else
        {
            b->addLine("// Comparing two objects of type " + ExtractCppTypename(typenavigator));
        }

        switch (typenavigator->getMaybe())
        {
            case type::Maybe::eNone:
            {
                b->appendBlock(block_if_data);
                break;
            }
            case type::Maybe::eOptional:
            {
                b->addLine("if (" + accessor + ".has_value() and " + otherInstanceAccessor + ".has_value()) // both have a value set");
                b->addBlock(block_if_data);
                b->addLine("if ((" + accessor + ".has_value() and !" + otherInstanceAccessor + ".has_value()) or (!" + accessor + ".has_value() and " + otherInstanceAccessor + ".has_value())) // only one has a value set (XOR)");
                b->addLine("\t return false;");
                break;
            }
            case type::Maybe::eRawPointer:
            {
                b->addLine("if (" + accessor + " != NULL and " + otherInstanceAccessor + " != NULL) // both have a value set");
                b->addBlock(block_if_data);
                b->addLine("if ((" + accessor + " != NULL and " + otherInstanceAccessor + " == NULL) or (" + accessor + " == NULL and " + otherInstanceAccessor + " != NULL)) // only one has a value set (XOR)");
                b->addLine("\t return false;");
                break;
            }
            case type::Maybe::eSharedPointer:
                [[fallthrough]];
            case type::Maybe::eUniquePointer:
            {
                b->addLine("if (" + accessor + " != nullptr and " + otherInstanceAccessor + " != nullptr) // both have a value set");
                b->addBlock(block_if_data);
                b->addLine("if ((" + accessor + " != nullptr and " + otherInstanceAccessor + " == nullptr) or (" + accessor + " == nullptr and " + otherInstanceAccessor + " != nullptr)) // only one has a value set (XOR)");
                b->addLine("\t return false;");
                break;
            }
            default:
            {
                throw error::MaybeNotValidException("Serializer", "ResolveMaybeEqualsBlock", "Unkown Maybe type (perhaps forgot to add in switch case?)", typenavigator->getMaybe(), typenavigator->getPath());
            }
        }
        return b;
    }

    CppBlockPtr Serializer::ResolveMaybeResetSoftBlock(const std::string& accessor, const CppBlockPtr& block_if_data, const typenavigator::NavigatorPtr& typenavigator)
    {
        CppBlockPtr b = std::make_shared<CppBlock>();

        if (!accessor.empty())
        {
            b->addLine("// Resetting soft member " + accessor);
        }
        else
        {
            b->addLine("// Ressetting soft the type " + ExtractCppTypename(typenavigator));
        }

        switch (typenavigator->getMaybe())
        {
            case type::Maybe::eNone:
            {
                b->appendBlock(block_if_data);
                break;
            }
            case type::Maybe::eOptional:
            {
                b->addLine("if (" + accessor + ".has_value())");
                b->addBlock(block_if_data);
                break;
            }
            case type::Maybe::eRawPointer:
            {
                b->addLine("if (" + accessor + " != NULL)");
                b->addBlock(block_if_data);
                break;
            }
            case type::Maybe::eSharedPointer:
                [[fallthrough]];
            case type::Maybe::eUniquePointer:
            {
                b->addLine("if (" + accessor + " != nullptr)");
                b->addBlock(block_if_data);
                break;
            }
            default:
            {
                throw error::MaybeNotValidException("Serializer", "ResolveMaybeResetSoftBlock", "Unkown Maybe type (perhaps forgot to add in switch case?)", typenavigator->getMaybe(), typenavigator->getPath());
            }
        }
        return b;
    }

    CppBlockPtr Serializer::ResolveMaybeResetHardBlock(const std::string& accessor, const CppBlockPtr& block_if_data, const typenavigator::NavigatorPtr& typenavigator)
    {
        CppBlockPtr b = std::make_shared<CppBlock>();
        if (!accessor.empty())
        {
            b->addLine("// Resetting hard member " + accessor);
        }
        else
        {
            b->addLine("// Ressetting hard the type " + ExtractCppTypename(typenavigator));
        }

        switch (typenavigator->getMaybe())
        {
            case type::Maybe::eNone:
            {
                b->appendBlock(block_if_data);
                break;
            }
            case type::Maybe::eOptional:
                [[fallthrough]];
            case type::Maybe::eRawPointer:
                [[fallthrough]];
            case type::Maybe::eSharedPointer:
                [[fallthrough]];
            case type::Maybe::eUniquePointer:
            {
                b->addLine(accessor + " = {};");
                break;
            }
            default:
            {
                throw error::MaybeNotValidException("Serializer", "ResolveMaybeResetSoftBlock", "Unkown Maybe type (perhaps forgot to add in switch case?)", typenavigator->getMaybe(), typenavigator->getPath());
            }
        }
        return b;
    }

    std::string Serializer::EscapeAccessor(const std::string& accessor)
    {
        std::string escaped_accessor = accessor;
        for (const auto& [key, value] : ESCAPE_ACCESSORS)
        {
            escaped_accessor = simox::alg::replace_all(escaped_accessor, key, value);
        }
        return escaped_accessor;
    }

    std::string Serializer::UnescapeAccessor(const std::string& accessor)
    {
        std::string unescaped_accessor = accessor;
        for (const auto& [key, value] : ESCAPE_ACCESSORS)
        {
            unescaped_accessor = simox::alg::replace_all(unescaped_accessor, value, key);
        }
        return unescaped_accessor;
    }

    std::string Serializer::ExtractCppTypename(const typenavigator::NavigatorPtr& n)
    {
        SerializerPtr cpp = Serializer::FromAronTypeNaviagtorPtr(n);
        return cpp->getCoreCppTypename();
    }

    std::vector<std::string> Serializer::ExtractCppTypenames(const std::vector<typenavigator::NavigatorPtr>& n)
    {
        std::vector<std::string> ret;
        for (const auto& typenav : n)
        {
            ret.push_back(Serializer::ExtractCppTypename(typenav));
        }
        return ret;
    }

    SerializerPtr Serializer::FromAronTypeNaviagtorPtr(const typenavigator::NavigatorPtr& n)
    {
        ARMARX_CHECK_NOT_NULL(n);
        return FACTORY->create(n, n->getPath());
    }

    // public methods
    std::string Serializer::getCoreCppTypename() const
    {
        return cppTypename;
    }

    std::string Serializer::getAronDataTypename() const
    {
        return aronDataTypename;
    }

    std::string Serializer::getAronDataPtrTypename() const
    {
        return getAronDataTypename() + "::PointerType";
    }

    std::string Serializer::getAronTypeTypename() const
    {
        return aronTypeTypename;
    }

    std::string Serializer::getAronTypePtrTypename() const
    {
        return getAronTypeTypename() + "::PointerType";
    }

    CppCtorPtr Serializer::toCtor(const std::string& name) const
    {
        CppCtorPtr c = CppCtorPtr(new CppCtor(name + "()"));
        std::vector<std::pair<std::string, std::string>> initList = this->getCtorInitializers("");
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("resetHard();");
        b->appendBlock(this->getCtorBlock(""));
        c->addInitListEntries(initList);
        c->setBlock(b);

        return c;
    }

    CppMethodPtr Serializer::toResetSoftMethod() const
    {
        std::stringstream doc;
        doc << "@brief resetSoft() - This method resets all member variables with respect to the current parameterization. \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void resetSoft() override", doc.str()));
        CppBlockPtr b = this->getResetSoftBlock("");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr Serializer::toResetHardMethod() const
    {
        std::stringstream doc;
        doc << "@brief resetHard() - This method resets member variables according to the XML type description. \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void resetHard() override", doc.str()));
        CppBlockPtr b = this->getResetHardBlock("");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr Serializer::toWriteTypeMethod() const
    {
        std::stringstream doc;
        doc << "@brief writeType() - This method returns a new type from the class structure using a type writer implementation. This function is static. \n";
        doc << "@return - the result of the writer implementation";

        CppMethodPtr m = CppMethodPtr(new CppMethod("static void writeType(armarx::aron::typeIO::WriterInterface& w, armarx::aron::type::Maybe __aronMaybeType = armarx::aron::type::Maybe::eNone)", doc.str()));
        CppBlockPtr b = this->getWriteTypeBlock("");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr Serializer::toWriteMethod() const
    {
        std::stringstream doc;
        doc << "@brief write() - This method returns a new type from the member data types using a data writer implementation. \n";
        doc << "@param w - The writer implementation\n";
        doc << "@return - the result of the writer implementation";

        CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void write(armarx::aron::dataIO::WriterInterface& w) const override", doc.str()));
        CppBlockPtr b = this->getWriteBlock("");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr Serializer::toReadMethod() const
    {
        std::stringstream doc;
        doc << "@brief read() - This method sets the struct members to new values given in a data reader implementation. \n";
        doc << "@param r - The reader implementation\n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void read(armarx::aron::dataIO::ReaderInterface& r, bool skip_first_readStartDict = false) override", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("this->resetSoft();");
        b->appendBlock(this->getReadBlock(""));
        m->setBlock(b);
        return m;
    }

    CppMethodPtr Serializer::toSpecializedDataWriterMethod(const std::string& returnname, const std::string& methodname, const std::string& writerName, const std::string& enforceConversion) const
    {
        std::stringstream doc;
        doc << "@brief specializedDataWrite() - This method returns a new data from the member data types using a writer implementation. \n";
        doc << "@return - the result of the writer implementation";

        CppMethodPtr m = CppMethodPtr(new CppMethod(returnname + " " + methodname + "() const", doc.str()));
        m->addLine(writerName + " writer;");
        m->addLine("this->write(writer);");
        m->addLine("return " + enforceConversion + "(writer.getResult());");
        return m;
    }

    CppMethodPtr Serializer::toSpecializedDataReaderMethod(const std::string& argumentname, const std::string& methodname, const std::string& readerName, const std::string& enforceConversion) const
    {
        std::stringstream doc;
        doc << "@brief specializedDataRead() - This method sets the struct members to new values given in a reader implementation. \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("void " + methodname + "(const " + argumentname + "& input)", doc.str()));
        m->addLine(readerName + " reader(" + enforceConversion + "(input));");
        m->addLine("this->read(reader);");
        return m;
    }

    CppMethodPtr Serializer::toSpecializedTypeWriterMethod(const std::string& returnname, const std::string& methodname, const std::string& writerName, const std::string& enforceConversion) const
    {
        std::stringstream doc;
        doc << "@brief specializedTypeWrite() - This method returns a new type from the member data types using a writer implementation. \n";
        doc << "@return - the result of the writer implementation";

        CppMethodPtr m = CppMethodPtr(new CppMethod("static " + returnname + " " + methodname + "()", doc.str()));
        m->addLine(writerName + " writer;");
        m->addLine("writeType(writer);");
        m->addLine("return " + enforceConversion + "(writer.getResult());");
        return m;
    }

    CppMethodPtr Serializer::toEqualsMethod() const
    {
        std::stringstream doc;
        doc << "@brief operator==() - This method checks whether all values equal another instance. \n";
        doc << "@param i - The other instance\n";
        doc << "@return - true, if all members are the same, false otherwise";

        CppMethodPtr m = CppMethodPtr(new CppMethod("bool operator==(const " + this->getFullCppTypename() + "& i) const", doc.str()));
        CppBlockPtr b = this->getEqualsBlock("", "i");
        b->addLine("return true;");
        m->setBlock(b);
        return m;
    }

}


