/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>

// Simox
#include <SimoxUtility/json.h>

// Base Class
#include <RobotAPI/libraries/aron/core/io/dataIO/reader/ReaderToken.h>

// ArmarX
#include <RobotAPI/libraries/aron/core/Config.h>
#include <RobotAPI/libraries/aron/core/Resolver.h>



namespace armarx::aron::dataIO::reader
{
    // BaseClass
    class NlohmannJSONReaderToken;
    typedef std::shared_ptr<NlohmannJSONReaderToken> NlohmannJSONReaderTokenPtr;

    class NlohmannJSONReaderToken :
        virtual public ReaderTokenInterface<nlohmann::json>
    {
    public:
        using PointerType = NlohmannJSONReaderTokenPtr;

        // constructors
        NlohmannJSONReaderToken() = delete;
        NlohmannJSONReaderToken(const nlohmann::json& data)
        {
            if (!data.is_object())
            {
                throw error::AronException("NlohmannJSONReaderToken", "NlohmannJSONReaderToken", "Received an invalid json instance for a reader. Allowed are only objects.");
            }

            descriptor = data::STRING_TO_DESCRIPTOR(data.at(io::Data::READER_WRITER_TYPE_SLUG));
            element = data;

            switch (descriptor)
            {
                case data::Descriptor::eDict:
                {
                    for (auto it = data.begin(); it != data.end(); ++it)
                    {
                        if (it.key() == io::Data::READER_WRITER_TYPE_SLUG || it.key() == io::Data::READER_WRITER_VERSION_SLUG)
                        {
                            continue;
                        }
                        allMemberNames.push_back(it.key());
                    }
                    childrenSize = allMemberNames.size();
                    break;
                }
                case data::Descriptor::eList:
                {
                    childrenSize = data.at(io::Data::READER_WRITER_VALUE_SLUG).size();
                    break;
                }
                default:
                    throw error::DescriptorNotValidException("NlohmannJSONReaderToken", "NlohmannJSONReaderToken", "Received an invalid instance for a reader token.", descriptor);
            }
        }

        data::Descriptor getTypeOfNext() const override
        {
            const nlohmann::json next = getNextElement();
            if (next.is_null())
            {
                return data::Descriptor::eUnknown;
            }

            std::string type_as_str = next.at(io::Data::READER_WRITER_TYPE_SLUG);
            data::Descriptor desc = data::STRING_TO_DESCRIPTOR(type_as_str);
            return desc;
        }

        nlohmann::json getNextElement() const override
        {
            if (!hasNextElement())
            {
                return nlohmann::json();
            }

            switch (descriptor)
            {
                case data::Descriptor::eDict:
                {
                    nlohmann::json ret = element.at(getCurrentKey());
                    return ret;
                }
                case data::Descriptor::eList:
                {
                    nlohmann::json ret = element.at(io::Data::READER_WRITER_VALUE_SLUG).at(currentIndex);
                    return ret;
                }
                default:
                    throw error::DescriptorNotValidException("NlohmannJSONReaderToken", "getNextElement", "Could not resove a type of a json. Allowed are only container types due to performance", descriptor);
            }
        }

        virtual bool hasNextElement() const override
        {
            switch (descriptor)
            {
                case data::Descriptor::eDict:
                {
                    nlohmann::json ret = element.at(getCurrentKey());
                    return !ret.is_null();
                }
                case data::Descriptor::eList:
                {
                    auto size = element.at(io::Data::READER_WRITER_VALUE_SLUG).size();
                    return size >= currentIndex;
                }
                default:
                    throw error::DescriptorNotValidException("NlohmannJSONReaderToken", "hasData", "Could not resove a type of a json. Allowed are only container types due to performance", descriptor);
            }
        }

    private:
    };
}
