/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Simox
#include <SimoxUtility/algorithm/vector.hpp>

// Header
#include "Visitor.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Resolver.h>



namespace armarx::aron::typeIO
{

    void Visitor::VisitAndSetup(WriterInterface& writer, const typenavigator::NavigatorPtr& aron)
    {
        VisitAndSetup(writer, aron->toAronPtr());
    }

    void Visitor::VisitAndSetup(WriterInterface& writer, const type::AronTypePtr& aron)
    {
        type::Descriptor desc = Resolver::GetDescriptor(aron);
        switch (desc)
        {
            case type::Descriptor::eObject:
            {
                auto casted = type::AronObjectPtr::dynamicCast(aron);
                writer.writeStartObject({casted->objectName, casted->maybe});
                for (const auto& [key, value] : casted->elementTypes)
                {
                    writer.writeKey(key);
                    Visitor::VisitAndSetup(writer, value);
                }
                writer.writeEndObject();
                break;
            }
            case type::Descriptor::eDict:
            {
                auto casted = type::AronDictPtr::dynamicCast(aron);
                writer.writeStartDict({casted->maybe});
                Visitor::VisitAndSetup(writer, casted->acceptedType);
                writer.writeEndDict();
                break;
            }
            case type::Descriptor::eTuple:
            {
                auto casted = type::AronTuplePtr::dynamicCast(aron);
                writer.writeStartTuple({casted->maybe});
                for (const auto& value : casted->elementTypes)
                {
                    Visitor::VisitAndSetup(writer, value);
                }
                writer.writeEndTuple();
                break;
            }
            case type::Descriptor::eList:
            {
                auto casted = type::AronListPtr::dynamicCast(aron);
                writer.writeStartList({casted->maybe});
                Visitor::VisitAndSetup(writer, casted->acceptedType);
                writer.writeEndList();
                break;
            }
            case type::Descriptor::eEigenMatrix:
            {
                auto casted = type::AronEigenMatrixPtr::dynamicCast(aron);
                writer.writeEigenMatrix({(unsigned int) casted->rows, (unsigned int) casted->cols, casted->typeName, casted->maybe});
                break;
            }
            case type::Descriptor::eEigenQuaternion:
            {
                auto casted = type::AronEigenQuaternionPtr::dynamicCast(aron);
                writer.writeEigenQuaternion({casted->typeName, casted->maybe});
                break;
            }
            case type::Descriptor::eIVTCByteImage:
            {
                auto casted = type::AronIVTCByteImagePtr::dynamicCast(aron);
                writer.writeIVTCByteImage({casted->maybe});
                break;
            }
            case type::Descriptor::eOpenCVMat:
            {
                auto casted = type::AronOpenCVMatPtr::dynamicCast(aron);
                writer.writeOpenCVMat({casted->maybe});
                break;
            }
            case type::Descriptor::ePCLPointCloud:
            {
                auto casted = type::AronPCLPointCloudPtr::dynamicCast(aron);
                writer.writePCLPointCloud({casted->typeName, casted->maybe});
                break;
            }
            case type::Descriptor::ePosition:
            {
                auto casted = type::AronPositionPtr::dynamicCast(aron);
                writer.writePosition({casted->maybe});
                break;
            }
            case type::Descriptor::eOrientation:
            {
                auto casted = type::AronOrientationPtr::dynamicCast(aron);
                writer.writeOrientation({casted->maybe});
                break;
            }
            case type::Descriptor::ePose:
            {
                auto casted = type::AronPosePtr::dynamicCast(aron);
                writer.writePose({casted->maybe});
                break;
            }
            case type::Descriptor::eInt:
            {
                auto casted = type::AronIntPtr::dynamicCast(aron);
                writer.writeInt({casted->maybe});
                break;
            }
            case type::Descriptor::eLong:
            {
                auto casted = type::AronLongPtr::dynamicCast(aron);
                writer.writeLong({casted->maybe});
                break;
            }
            case type::Descriptor::eFloat:
            {
                auto casted = type::AronFloatPtr::dynamicCast(aron);
                writer.writeFloat({casted->maybe});
                break;
            }
            case type::Descriptor::eDouble:
            {
                auto casted = type::AronDoublePtr::dynamicCast(aron);
                writer.writeDouble({casted->maybe});
                break;
            }
            case type::Descriptor::eString:
            {
                auto casted = type::AronStringPtr::dynamicCast(aron);
                writer.writeString({casted->maybe});
                break;
            }
            case type::Descriptor::eBool:
            {
                auto casted = type::AronBoolPtr::dynamicCast(aron);
                writer.writeBool({casted->maybe});
                break;
            }
            case type::Descriptor::eTime:
            {
                auto casted = type::AronTimePtr::dynamicCast(aron);
                writer.writeTime({casted->maybe});
                break;
            }
            default:
            {
                throw error::DescriptorNotValidException("Visitor", "VisitAndSetup", "Type could not be resolved. The ice_id of the input was: " + aron->ice_id(), desc);
            }
        }
    }
}
