/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "EigenMatrix.h"

// Simox
#include <SimoxUtility/algorithm/vector.hpp>

namespace armarx::aron::typenavigator
{
    // constructors
    EigenMatrixNavigator::EigenMatrixNavigator(const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eEigenMatrix, path)
    {
    }

    EigenMatrixNavigator::EigenMatrixNavigator(const type::AronEigenMatrixPtr& o, const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eDict, path),
        detail::NavigatorBase<type::AronEigenMatrix, EigenMatrixNavigator>(o)
    {
    }

    // operators
    bool EigenMatrixNavigator::operator==(const EigenMatrixNavigator& other) const
    {
        if (getMaybe() != other.getMaybe())
        {
            return false;
        }
        return getCols() == other.getCols() && getRows() == other.getRows() && getTypename() == other.getTypename();
    }

    std::string EigenMatrixNavigator::checkTypename(const std::string& s) const
    {
        for (const auto& [key, list] : ACCEPTED_TYPES)
        {
            if (s == key)
            {
                return key;
            }

            for (const auto& el : list)
            {
                if (s == el)
                {
                    return key;
                }
            }
        }
        throw error::StringNotValidException("EigenMatrixNavigator", "checkTypename", "The typename is wrong.", s, getPath());
    }

    unsigned int EigenMatrixNavigator::getRows() const
    {
        return this->aron->rows;
    }

    unsigned int EigenMatrixNavigator::getCols() const
    {
        return this->aron->cols;
    }

    void EigenMatrixNavigator::setRows(int w)
    {
        if (w == 0 || w < -1)
        {
            throw error::AronException("AronEigenMatrixNavigator", "setRows", "The rows cannot be 0 or < -1", getPath());
        }
        this->aron->rows = w;
    }

    void EigenMatrixNavigator::setCols(int h)
    {
        if (h == 0 || h < -1)
        {
            throw error::AronException("AronEigenMatrixNavigator", "setCols", "The cols cannot be 0 or < -1", getPath());
        }
        this->aron->cols = h;
    }

    std::string EigenMatrixNavigator::getTypename() const
    {
        return this->aron->typeName;
    }

    void EigenMatrixNavigator::setTypename(const std::string& u)
    {
        this->aron->typeName = checkTypename(u);;
    }

    type::AronEigenMatrixPtr EigenMatrixNavigator::toAronEigenMatrixPtr() const
    {
        return this->aron;
    }

    // virtual implementations
    std::string EigenMatrixNavigator::getName() const
    {
        return "AronEigenMatrix<" + std::to_string(this->aron->rows) + ", " + std::to_string(this->aron->cols) + ", " + this->aron->typeName + ">";
    }
}

